/* 

    TiMidity -- Experimental MIDI to WAVE converter
    Copyright (C) 1995 Tuukka Toivonen <titoivon@snakemail.hut.fi>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   audio.c

   Audio output (to file / device) functions.
   */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <linux/soundcard.h>

#include "config.h"
#include "common.h"
#include "instruments.h"
#include "playmidi.h"
#include "controls.h"
#include "audio.h"

static char *orig_RIFFheader=
  "RIFF" "\377\377\377\377" 
  "WAVE" "fmt " "\020\000\000\000" "\001\000"
  /* 22: channels */ "\001\000" 
  /* 24: freq */ "xxxx" 
  /* 28: bytes/sec */ "xxxx" 
  /* 32: bytes/sample */ "\004\000" 
  /* 34: bits/sample */ "\020\000"
  "data" "\377\377\377\377"
;

int output_rate=DEFAULT_RATE, play_mode=PLAY_DEVICE, mono, eight_bit;
int buffer_fragments=-1;
char *audio_name=AUDIO_DEVICE;

static int audio_device;

static void adjust_to_sample_rate(void)
{
  if (!control_ratio)
    {
      control_ratio = output_rate / CONTROLS_PER_SECOND;
      if(control_ratio<1)
	control_ratio=1;
      else if (control_ratio > MAX_CONTROL_RATIO)
	control_ratio=MAX_CONTROL_RATIO;
    }
}

/* Makes the audio device play all buffered data before returning.
   This shouldn't leave the audio device closed, or its parameters
   altered. */
void flush_output(void)
{
  if (play_mode==PLAY_DEVICE)
    ioctl(audio_device, SNDCTL_DSP_SYNC);
}

/* Makes the audio device purge all buffered data and stop playing
   immediately. This shouldn't leave the audio device closed, or its
   parameters altered. */
void purge_output(void)
{
  if (play_mode==PLAY_DEVICE)
    ioctl(audio_device, SNDCTL_DSP_RESET);
}

int open_output(void)
{
  int i,j,fd=0; /* avoid silly warning about uninitialized fd */
  if (play_mode==PLAY_DEVICE)
    {
      fd=open(audio_name, O_RDWR | O_NDELAY);
      if (fd<0)
	{
	  cmsg(CMSG_ERROR, VERB_NORMAL, "%s: %s",
	       audio_name, sys_errlist[errno]);
	  return -1;
	}

      j=i=(eight_bit) ? 8 : 16;
      if ((ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &i)<0) || i!=j)
	{
	  i=j=8;
	  if (eight_bit || (ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &i)<0) || i!=j)
	    {
	      cmsg(CMSG_ERROR, VERB_NORMAL, 
		   "%s doesn't support %s8-bit sample width",
		   audio_name, (eight_bit) ? "" : "16- or ");
	      close(fd);
	      return -1;
	    }
	  cmsg(CMSG_WARNING, VERB_VERBOSE, 
	       "Sample width adjusted to 8 bits");
	  eight_bit=1; /* we've got 8-bit samples */
	}

      i=j=(mono) ? 0 : 1;
      if ((ioctl(fd, SNDCTL_DSP_STEREO, &i)<0) || i!=j)
	{
	  i=j=0;
	  if (mono || (ioctl(fd, SNDCTL_DSP_STEREO, &i)<0) || i!=j)
	    {
	      cmsg(CMSG_ERROR, VERB_NORMAL, 
		   "%s doesn't support %smono samples",
		   audio_name, (mono) ? "" : "stereo or ");
	      close(fd);
	      return -1;
	    }
	  mono=1;
	  cmsg(CMSG_WARNING, VERB_VERBOSE, "Sound adjusted to monophonic");
	}

      i=output_rate;
      if (ioctl(fd, SNDCTL_DSP_SPEED, &i)<0)
	{
	  cmsg(CMSG_ERROR, VERB_NORMAL, 
	       "%s doesn't support a %d Hz sample rate",
	       audio_name, output_rate);
	  close(fd);
	  return -1;	  
	}

      if (i != output_rate)
	{
	  output_rate=i;
	  cmsg(CMSG_WARNING, VERB_VERBOSE,
	       "Output rate adjusted to %d Hz", output_rate);
	}

      i=AUDIO_BUFFER_BITS;
      if (!mono) i++;
      if (!eight_bit) i++;

      if (buffer_fragments==-1)
	buffer_fragments=0;

      if (interactive)
	i |= (buffer_fragments<<16);
      j=i;
      if (ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &i)<0)
	{
	  cmsg(CMSG_WARNING, VERB_NORMAL, 
	       "%s doesn't like %d-byte buffer fragments", 
		  audio_name, (1<<j));
	  /* It should still work in some fashion.
	     We should set up our own reserve buffer anyway -- 64k
	     isn't enough. */
	}
      adjust_to_sample_rate();
      audio_device=fd;
      return 0;
    }
  else if (play_mode==PLAY_FILE || play_mode==PLAY_WAVE)
    {
      if (audio_name && *audio_name && strcmp(audio_name, "-"))
	{
	  fd=open(audio_name, O_CREAT|O_TRUNC|O_WRONLY, 0644);
	  if (fd<0)
	    {
	      cmsg(CMSG_ERROR, VERB_NORMAL, "%s: %s",
		   audio_name, sys_errlist[errno]);
	      return -1;
	    }
	}
      else
	{
	  fd=1; /* stdout */
	  msgfp = stderr; /* and messages to stderr */
	}

      adjust_to_sample_rate();
      audio_device=fd;
      if (play_mode==PLAY_WAVE)
	{
	  /* Output a (probably non-standard) RIFF header */
	  char RIFFheader[44];
	  int t;
	  memcpy(RIFFheader, orig_RIFFheader, 44);
	  if (!mono) RIFFheader[22]='\002';
	  *((int *)(RIFFheader+24))=LE_LONG(output_rate);
	  t=output_rate; if (!mono) t*=2; if (!eight_bit) t*=2;
	  *((int *)(RIFFheader+28))=LE_LONG(t);
	  if (mono && eight_bit) RIFFheader[32]='\001';
	  else if (mono || eight_bit) RIFFheader[32]='\002';
	  if (eight_bit) RIFFheader[34]='\010';
	  write(fd, RIFFheader, 44);
	}
      return 0;
    }
  else 
    return -1;
}

void close_output(void)
{
  if (play_mode==PLAY_DEVICE)
    close(audio_device);
  else if (play_mode==PLAY_FILE || play_mode==PLAY_WAVE)
    {
      if (audio_device != 1) /* We may be playing on stdout */
	close(audio_device);
    }
}

/*
   Gets an array of ints which contain signed data which should be
   scaled to whatever the output device or file accepts and written 
   out. If the global variable mono is not set, then each sample is
   a stereo pair (left,right).
*/
void output_data(int *buf, int count)
{
  int c;
  int *lp=buf, l,n;
  c=count;
  if (!mono) c *= 2;
  if (eight_bit)
    {
      char *cp=(char *)(lp);
      while (c--)
	{
	  l=(*lp++)>>22;
	  if (l>127) l=127;
	  else if (l<-128) l=-128;
	  *cp++ = 0x80 ^ l;
	}
      do
	n=write(audio_device, buf, (mono) ? count : count * 2);
      while (n==-1 && errno==EINTR);
    }
  else
    {
      short *sp=(short *)(lp);
      while (c--)
	{
	  l=(*lp++) >> 14;
	  if (l > 32767) l=32767;
	  else if (l<-32768) l=-32768;
	  *sp++ = l;
	}
      do
	n=write(audio_device, buf, (mono) ? count * 2 : count*4);
      while (n==-1 && errno==EINTR);
    }
}
