/*

    TiMidity -- Experimental MIDI to WAVE converter
    Copyright (C) 1995 Tuukka Toivonen <titoivon@snakemail.hut.fi>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   common.c

   */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include "config.h"
#include "common.h"
#include "audio.h"
#include "controls.h"

/* Some global variables */

FILE *msgfp=stdout, *inputfp=stdin;
char *program_name, current_filename[1024];

/* Local variables */

#ifdef DEFAULT_PATH
    /* The paths in this list will be tried whenever we're reading a file */
    static PathList defaultpathlist={DEFAULT_PATH,0};
    static PathList *pathlist=&defaultpathlist; /* This is a linked list */
#else
    static PathList *pathlist=0;
#endif

/* Should print out a real message... when curses goes, it clears the
   screen and you have no idea why we bailed out */
void bail_out(int c)
{
  static int bailing_out=0;
  if (bailing_out)
    {
      fprintf(stderr, "Recursive call to bail_out: %d\nExiting", c);
      exit(c);
    }
  bailing_out=1;
  purge_output();
  close_output();
#ifdef INTERACTIVE
  close_controls();
#endif
  fprintf(stderr, "Bailing out: %d\n", c);
  exit(c);
}

/* Functions */

void add_to_pathlist(char *s)
{
  PathList *plp=safe_malloc(sizeof(PathList));
  strcpy((plp->path=safe_malloc(strlen(s)+1)),s);
  plp->next=pathlist;
  pathlist=plp;
}

/* Try to open a file for reading. If the filename ends in one of the 
   defined compressor extensions, pipe the file through the decompressor */
static FILE *try_to_open(char *name, int decompress, int noise_mode)
{
  FILE *fp;

  fp=fopen(name, "r"); /* First just check that the file exists */

  if (!fp)
    return 0;

#ifdef DECOMPRESSOR_LIST
  if (decompress)
    {
      int l,el;
      static char *decompressor_list[] = DECOMPRESSOR_LIST, **dec;
      char tmp[512];
      /* Check if it's a compressed file */ 
      l=strlen(name);
      for (dec=decompressor_list; *dec; dec+=2)
	{
	  el=strlen(*dec);
	  if ((el>=l) || (strcmp(name+l-el, *dec)))
	    continue;

	  /* Yes. Close the file, open a pipe instead. */
	  fclose(fp);
	  sprintf(tmp, *(dec+1), name);
	  fp=popen(tmp, "r");
	  break;
	}
    }
#endif
  
  return fp;
}

void close_file(FILE *fp)
{
#ifdef DECOMPRESSOR_LIST
  if (pclose(fp)) /* Any better ideas? */
#endif
    fclose(fp);
}

/* This is meant to find and open files for reading, possibly piping
   them through a decompressor. */
FILE *open_file(char *name, int decompress, int noise_mode)
{
  FILE *fp;
  PathList *plp;
  int l;

  if (!name || !(*name))
    {
      cmsg(CMSG_ERROR, VERB_NORMAL, "Attempted to open nameless file.");
      return 0;
    }

  /* First try the given name */

  plp=pathlist;
  strncpy(current_filename, name, 1023);
  current_filename[1023]='\0';

  cmsg(CMSG_INFO, VERB_DEBUG, "Trying to open %s", current_filename);
  if ((fp=try_to_open(current_filename, decompress, noise_mode)))
    return fp;

  if (noise_mode && (errno != ENOENT))
    {
      cmsg(CMSG_ERROR, VERB_NORMAL, "%s: %s", 
	   current_filename, sys_errlist[errno]);
      return 0;
    }
  
  if (name[0] != '/')
    while (plp)  /* Try along the path then */
      {
	*current_filename=0;
	l=strlen(plp->path);
	if(l)
	  {
	    strcpy(current_filename, plp->path);
	    if(current_filename[l-1]!='/')
	      strcat(current_filename, "/");
	  }
	strcat(current_filename, name);
	cmsg(CMSG_INFO, VERB_DEBUG, "Trying to open %s", current_filename);
	if ((fp=try_to_open(current_filename, decompress, noise_mode)))
	  return fp;
	if (noise_mode && (errno != ENOENT))
	  {
	    cmsg(CMSG_ERROR, VERB_NORMAL, "%s: %s", 
		 current_filename, sys_errlist[errno]);
	    return 0;
	  }
	plp=plp->next;
      }
  
  /* Nothing could be opened. */

  *current_filename=0;
  
  if (noise_mode>=2)
    cmsg(CMSG_ERROR, VERB_NORMAL, "%s: %s", name, sys_errlist[errno]);
  
  return 0;
}

void *safe_malloc(int count)
{
  void *p;
  if (count < 0 || count > (1<<21))
    {
      cmsg(CMSG_FATAL, VERB_NORMAL, 
	   "Strange, I feel like allocating %d bytes. This must be a bug.",
	   count);
      bail_out(2);
    }

  if (!(p=malloc(count)))
    {
      cmsg(CMSG_FATAL, VERB_NORMAL, 
	   "Sorry. Couldn't malloc %d bytes.", count);
      bail_out(2);
    }
  return p;
}

/* This is meant for skipping a few bytes in a file or fifo. */
void skip(FILE *fp, int len)
{
  int c;
  char tmp[1024];
  while (len>0)
    {
      c=len;
      if (c>1024) c=1024;
      len-=c;
      if (c!=fread(tmp, 1, c, fp))
	cmsg(CMSG_ERROR, VERB_NORMAL, "%s: skip: %s", 
	     current_filename, sys_errlist[errno]);
    }
}
