/* 

    TiMidity -- Experimental MIDI to WAVE converter
    Copyright (C) 1995 Tuukka Toivonen <titoivon@snakemail.hut.fi>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   controls.c
   
   */

#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <stdarg.h>

#ifdef IA_NCURSES
#include <ncurses.h>
#endif

#include "config.h"
#include "common.h"
#include "audio.h"
#include "instruments.h"
#include "playmidi.h"
#include "controls.h"

int verbosity=0, interactive=0, trace_playing=0;

/***********************************************************************/
#ifdef IA_NCURSES
/* foreground/background checks disabled since switching to curses */
static int in_foreground=1, on_curses=0;

static int ctl_helpmode=0;

void ctl_refresh(void)
{
  move(0,0);
  refresh();
}

static void ctl_help_mode(void)
{
  static WINDOW *helpwin;
  if (ctl_helpmode)
    {
      ctl_helpmode=0;
      delwin(helpwin);
      touchwin(stdscr);
      ctl_refresh();
    }
  else
    {
      ctl_helpmode=1;
      /* And here I thought the point of curses was that you could put
	 stuff on windows without having to worry about whether this
	 one is overlapping that or the other way round... */
      helpwin=newwin(2,COLS,0,0);
      wattron(helpwin, A_REVERSE);
      wclear(helpwin);
      waddstr(helpwin, 
 "  V=Louder volume   b=Skip back      n=Next file       r=Restart file");
      wmove(helpwin, 1,0);
      waddstr(helpwin, 
 "  v=Softer volume   f=Skip forward   p=Previous file   q=Quit program");
      wattroff(helpwin, A_REVERSE);
      wrefresh(helpwin);
    }
}

static void ctl_labels(void)
{
  int i;
  clear();
  move(0,0);
  addstr("TiMidity v" TIMID_VERSION);
  move(0,COLS-52);
  addstr("(C) 1995 Tuukka Toivonen <titoivon@snakemail.hut.fi>");
  move(1,0);
  addstr("Press 'h' for help with keys, or 'q' to quit.");
  move(3,0);
  addstr("File:");
  move(4,0);
  addstr("Time:");
  move(4,6+6+1);
  addch('/');
  move(4,40);
  printw("Voices:    / %d", voices);
  move(4,COLS-20);
  addstr("Master volume:");
  move(5,0);
  for (i=0; i<COLS; i++)
    addch('_');
  move(6,0);
  addstr("Ch");
  move(6,COLS-20);
  addstr("Prg Vol Exp Pan S B");
  move(7,0);
  for (i=0; i<COLS; i++)
    addch('-');
  for (i=0; i<16; i++)
    {
      move(8+i, 0);
      printw("%02d", i+1);
    }
  ctl_refresh();
}

void ctl_total_time(int tt)
{
  int mins, secs=tt/output_rate;
  mins=secs/60;
  secs-=mins*60;
  move(4,6+6+3);
  attron(A_BOLD);
  printw("%3d:%02d", mins, secs);
  attroff(A_BOLD);
  ctl_refresh();
}

void ctl_master_volume(int mv)
{
  move(4,COLS-5);
  attron(A_BOLD);
  printw("%03d %%", mv);
  attroff(A_BOLD);
}

void ctl_file_name(char *name)
{
  move(3,6);
  clrtoeol();
  attron(A_BOLD);
  addstr(name);
  attroff(A_BOLD);
  ctl_refresh();
}

void ctl_current_time(int ct)
{
  int i,v;
  int mins, secs=ct/output_rate;
  mins=secs/60;
  secs-=mins*60;
  move(4,6);
  attron(A_BOLD);
  printw("%3d:%02d", mins, secs);
  v=0;
  i=voices;
  while (i--)
    if (voice[i].status!=VOICE_FREE) v++;
  move(4,48);
  printw("%2d", v);
  attroff(A_BOLD);
  ctl_refresh();
}

void ctl_reset(void)
{
  int i,j;
  for (i=0; i<16; i++)
    {
      move(8+i, 3);
      for (j=0; j<COLS-24; j++)
	addch('.');
      ctl_program(i, channel[i].program);
      ctl_volume(i, channel[i].volume);
      ctl_expression(i, channel[i].expression);
      ctl_panning(i, channel[i].panning);
      ctl_sustain(i, channel[i].sustain);
      ctl_pitch_bend(i, channel[i].pitchbend);
    }
  ctl_refresh();
}

void ctl_note(int v)
{
  int xl=voice[v].note%(COLS-24);
  move(8+voice[v].channel,xl+3);
  switch(voice[v].status)
    {
    case VOICE_FREE: addch('.'); break;
    case VOICE_ON: addch('0'+(10*voice[v].velocity)/128); break;
    case VOICE_OFF: addch(':'); break;
    case VOICE_SUSTAINED: addch('+'); break;
    }
}

void ctl_program(int channel, int val)
{
  move(8+channel, COLS-20);
  printw("%03d", val);
}

void ctl_volume(int channel, int val)
{
  move(8+channel, COLS-16);
  printw("%03d", val);  
}

void ctl_expression(int channel, int val)
{
  move(8+channel, COLS-12);
  printw("%03d", val);
}

void ctl_panning(int channel, int val)
{
  move(8+channel, COLS-8);
  if (val==NO_PANNING)
    addstr("   ");
  else if (val<5)
    addstr(" L ");
  else if (val>123)
    addstr(" R ");
  else if (val>60 && val<68)
    addstr(" C ");
  else
    {
      /* printw("%+02d", (100*(val-64))/64); */
      val = (100*(val-64))/64; /* piss on curses */
      if (val<0)
	{
	  addch('-');
	  val=-val;
	}
      else addch('+');
      printw("%02d", val);
    }
}

void ctl_sustain(int channel, int val)
{
  move(8+channel, COLS-4);
  if (val) addch('S');
  else addch(' ');
}

void ctl_pitch_bend(int channel, int val)
{
  move(8+channel, COLS-2);
  if (val>0x2000) addch('+');
  else if (val<0x2000) addch('-');
  else addch(' ');
}

/***********************************************************************/

int open_controls(void)
{
  if (interactive)
    {
      initscr();
      cbreak();
      noecho();
      nonl();
      nodelay(stdscr, 1);
      scrollok(stdscr, 1);
      idlok(stdscr, 1);
      on_curses=1;
      ctl_labels();
   }
  return 0;
}

int close_controls(void)
{
  if (interactive)
    {
      endwin();
      on_curses=0;
    }
  return 0;
}

int read_controls(int *valp)
{
  int c;
  if (!interactive || !in_foreground || !on_curses) 
    return RC_NONE;

  while ((c=getch())!=ERR)
    {
      switch(c)
	{
	case 'h':
	case '?':
	  ctl_help_mode();
	  return RC_NONE;

	case 'V':
	  *valp=10;
	  return RC_CHANGE_VOLUME;
	case 'v':
	  *valp=-10;
	  return RC_CHANGE_VOLUME;
	case 'q':
	  return RC_QUIT;
	case 'n':
	  return RC_NEXT;
	case 'p':
	  return RC_REALLY_PREVIOUS;
	case 'r':
	  return RC_RESTART;
	case 'f':
	  *valp=output_rate;
	  return RC_FORWARD;
	case 'b':
	  *valp=output_rate;
	  return RC_BACK;
	  /* case ' ':
	     return RC_TOGGLE_PAUSE; */
	}
    }
  return RC_NONE;
}
#else
void ctl_refresh(void){}
void ctl_reset(void){}
void ctl_note(int v){}
void ctl_total_time(int tt){}
void ctl_current_time(int ct){}
void ctl_file_name(char *name){}
void ctl_master_volume(int mv){}
void ctl_program(int channel, int val){}
void ctl_volume(int channel, int val){}
void ctl_expression(int channel, int val){}
void ctl_panning(int channel, int val){}
void ctl_sustain(int channel, int val){}
void ctl_pitch_bend(int channel, int val){}
#endif

int cmsg(int type, int verbosity_level, char *fmt, ...)
{
  va_list ap;
  if (verbosity<verbosity_level)
    return 0;
  va_start(ap, fmt);
#ifdef IA_NCURSES
  if (interactive && on_curses && in_foreground)
    {
      switch(type)
	{
	  /* Pretty pointless to only one line for messages */
	case CMSG_WARNING:
	case CMSG_ERROR:
	case CMSG_FATAL:
	  move(2,0);
	  clrtoeol();
	  attron(A_REVERSE);
	  vwprintw(stdscr, fmt, ap);
	  attroff(A_REVERSE);
	  ctl_refresh();
	  if (type==CMSG_WARNING)
	    sleep(1); /* Don't you just _HATE_ it when programs do this... */
	  else
	    sleep(2);
	  move(2,0);
	  clrtoeol();
	  ctl_refresh();
	  break;
	}
    }
  else
#endif
    {
      switch(type)
	{
	case CMSG_FATAL:
	case CMSG_ERROR:
	case CMSG_WARNING:
	  vfprintf(stderr, fmt, ap);
	  fprintf(stderr, "\n");
	  break;
	case CMSG_INFO:
	case CMSG_TEXT:
	  vfprintf(msgfp, fmt, ap);
	  fprintf(msgfp, "\n");
	  break;
	}
    }
  va_end(ap);
  return 0;
}
