/*

    TiMidity -- Experimental MIDI to WAVE converter
    Copyright (C) 1995 Tuukka Toivonen <titoivon@snakemail.hut.fi>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   instruments.c 
   
   Code to load and unload GUS-compatible instrument patches.

*/

#include <stdio.h>
#include <strings.h>
#include <malloc.h>

#include "config.h"
#include "common.h"
#include "instruments.h"
#include "playmidi.h"
#include "audio.h"
#include "controls.h"

/* Some functions get aggravated if not even the standard banks are 
   available. */
static ToneBank standard_tonebank, standard_drumset;
ToneBank 
  *tonebank[128]={&standard_tonebank}, 
  *drumset[128]={&standard_drumset};

/* This is a special instrument that can be specified on the command line 
   as the default one to use */
Instrument *default_instrument=0; 

int default_program=DEFAULT_PROGRAM;

static void free_instrument(Instrument *ip)
{
  Sample *sp;
  int i;
  for (i=0; i<ip->samples; i++)
    {
      sp=&(ip->sample[i]);
      free(sp->data);
    }
  free(ip->sample);
  free(ip);
}

static void free_bank(int dr, int b)
{
  int i;
  ToneBank *bank=((dr) ? drumset[b] : tonebank[b]);
  for (i=0; i<128; i++)
    if (bank->tone[i].instrument)
      {
	/* Not that this could ever happen, of course */
	if (bank->tone[i].instrument != MAGIC_LOAD_INSTRUMENT)
	  free_instrument(bank->tone[i].instrument);
	bank->tone[i].instrument=0;
      }
}

static int convert_envelope_rate(unsigned char rate)
{
  int r;
  r=3-((rate>>6) & 0x3);
  r*=3;
  r = (rate & 0x3f) << r; /* 6.9 fixed point */

  return (((r * 44100) / output_rate) * control_ratio) << 10;
  /* 15.15 fixed point. The scaling is estimated. I'm not sure how the
     GUS behaves. <<10 may be twice too fast, but it sounds reasonable
     with the exponential decay kludge. It also saves ~25% CPU against 
     <<9. */
}

static int convert_envelope_offset(unsigned char offset)
{
  /* 15.15 fixed point */
  return offset << (7+15);

  /* This is no good... Can anyone tell me what these values mean?
     Are they GUS-style "exponential" volumes? And what does that mean? 
     int exponent, mantissa;
     exponent = (offset>>4) & 0x0F;
     mantissa = offset & 0x0F;
     return (mantissa + (1 << (exponent+12));*/
}

static Instrument *load_instrument(char *name)
{
  Instrument *ip;
  Sample *sp;
  FILE *fp;
  char tmp[1024];
  int i,j;
  
  if (!name) return 0;
  
  if (!(fp=open_file(name, 1, OF_NORMAL)))
    {
      strncpy(tmp, name, 1019);
      tmp[1019]='\0';
      strcat(tmp, ".pat");
      if (!(fp=open_file(tmp, 1, OF_NORMAL)))
	{
	  cmsg(CMSG_ERROR, VERB_NORMAL, 
	       "Instrument `%s' can't be found.", name);
	  return 0;
	}
    }

  cmsg(CMSG_INFO, VERB_NOISY, "Loading instrument %s", current_filename);
  
  /* Read some headers and do cursory sanity checks. There are loads
     of magic offsets. This should be rewritten... */

  if ((239 != fread(tmp, 1, 239, fp)) ||
      memcmp(tmp, "GF1PATCH110\0ID#000002", 22))
    {
      cmsg(CMSG_ERROR, VERB_NORMAL, "%s: not an instrument", name);
      return 0;
    }
  
  if (tmp[82] != 1) /* instruments */
    {
      cmsg(CMSG_ERROR, VERB_NORMAL, 
	   "Can't handle patches with %d instruments", 
	   *((unsigned char *)(tmp) + 151));
      return 0;
    }

  if (tmp[151] != 1) /* layers. What's a layer? */
    {
      cmsg(CMSG_ERROR, VERB_NORMAL, 
	   "Can't handle instruments with %d layers", tmp[151]);
      return 0;
    }
  
  ip=safe_malloc(sizeof(Instrument));
  ip->samples = tmp[198];
  ip->sample = safe_malloc(sizeof(Sample) * ip->samples);
  for (i=0; i<ip->samples; i++)
    {
      long tmplong;
      unsigned short tmpshort;
      unsigned char fractions;
      
#define READ_CHAR(thing) if (1 != fread(&(thing), 1, 1, fp)) goto fail;
#define READ_SHORT(thing) if (1 != fread(&tmpshort, 2, 1, fp)) goto fail; \
      thing = LE_SHORT(tmpshort);
#define READ_LONG(thing) if (1 != fread(&tmplong, 4, 1, fp)) goto fail; \
      thing = LE_LONG(tmplong);

      skip(fp, 7); /* Skip the wave name */

      if (1 != fread(&fractions, 1, 1, fp))
	{
	fail:
	  cmsg(CMSG_ERROR, VERB_NORMAL, "Error reading sample %d", i);
	  for (j=0; j<i; j++)
	    free(ip->sample[j].data);
	  free(ip->sample);
	  free(ip);
	  return 0;
	}

      sp=&(ip->sample[i]);
      
      READ_LONG(sp->data_length);
      READ_LONG(sp->loop_start);
      READ_LONG(sp->loop_end);
      READ_SHORT(sp->sample_rate);
      READ_LONG(sp->low_freq);
      READ_LONG(sp->high_freq);
      READ_LONG(sp->root_freq);
      skip(fp, 2); /* Why have a "root frequency" and then "tuning"?? */
      
      READ_CHAR(tmp[0]);
      sp->panning = (tmp[0] * 8 + 4) & 0x7f;

      if (12 != fread(tmp, 1, 12, fp)) goto fail; /* envelope */

      READ_CHAR(sp->tremolo_sweep);
      READ_CHAR(sp->tremolo_rate);
      READ_CHAR(sp->tremolo_depth);
      READ_CHAR(sp->vibrato_sweep);
      READ_CHAR(sp->vibrato_rate);
      READ_CHAR(sp->vibrato_depth);

      READ_CHAR(sp->modes);

      skip(fp, 40); /* skip scale frequency, scale factor, reserved space */

      /* Do some processing */

      /* Somebody makes strange patches. This hack makes things even
	 stranger. Ideally, we'd do everything like the GUS sequencers
	 do. In practice, we guess at what they might do and then
	 approximate that. Also, a GUS may have time to process any
	 number of envelopes, but we'd like to skip the pointless ones. */

      /* Envelope rates all maxed out? Envelope ends with a high "offset"?
	 That's weird. We don't like weird. */
      if (!memcmp(tmp, "??????", 6) || 
	  *((unsigned char *)(tmp+11)) >= 100) 
	{
	  sp->modes &= ~(MODES_ENVELOPE);
	  cmsg(CMSG_INFO, VERB_DEBUG, 
	       " - Weirdness, removing envelope");
	}
      else if (!(sp->modes & MODES_SUSTAIN))
	{
	  /* Don't know if this is justified, but patches without sustain
	     usually don't need the envelope either. They're mostly drums. 
	     I think. */
	  sp->modes &= ~(MODES_ENVELOPE);
	  cmsg(CMSG_INFO, VERB_DEBUG, 
	       " - No sustain, removing envelope");
	}
      
      for (j=0; j<6; j++)
	{
	  sp->envelope_rate[j]=
	    convert_envelope_rate(tmp[j]);
	  sp->envelope_offset[j]= 
	    convert_envelope_offset(tmp[6+j]);
	}

      /* Then read the sample data */
      sp->data = safe_malloc(sp->data_length);
      if (1 != fread(sp->data, sp->data_length, 1, fp))
	goto fail;
      
      if (!(sp->modes & MODES_16BIT)) /* resample_data wants 16-bit data */
	{
	  int i=sp->data_length;
	  char *cp=(char *)(sp->data);
	  short *tmp,*new=safe_malloc(sp->data_length*2);
	  tmp=new;
	  while (i--)
	    *tmp++ = (*cp++) * 256;
	  cp=(char *)(sp->data);
	  sp->data = new;
	  free(cp);
	  sp->data_length *= 2;
	  sp->loop_start *= 2;
	  sp->loop_end *= 2;
	}
      if (sp->modes & MODES_UNSIGNED) /* Everybody wants signed data */
	{
	  int i=sp->data_length/2;
	  short *tmp=sp->data;
	  while (i--)
	    *tmp++ ^= 0x8000;
	}

      /* Try to determine a volume scaling factor for the sample.
	 I don't know why this should be necessary, but even though
	 this is a very crude adjustment, things sound much better and
	 more balanced with it. Still, this should be a runtime option. */
      {
	int i=sp->data_length/2;
	int maxamp=0,a;
	short *tmp=sp->data;
	while (i--)
	  {
	    a=*tmp++;
	    if (a<0) a=-a;
	    if (a>maxamp)
	      maxamp=a;
	  }
	sp->volume=(1<<30)/maxamp;
	cmsg(CMSG_INFO, VERB_DEBUG, " * volume comp: %d", sp->volume);
      }
      
      sp->data_length /= 2; /* The low bits shouldn't ever be set, */
      sp->loop_start /= 2;  /* I suppose. Now they won't be. */
      sp->loop_end /= 2;

      /* These to be specified in units of fractional samples */
      sp->data_length <<= FRACTION_BITS;
      sp->loop_start <<= FRACTION_BITS;
      sp->loop_end <<= FRACTION_BITS;

      /* Adjust for fractional loop points. This is a guess. Does anyone
	 know what "fractions" really stands for? */
      sp->loop_start |=
	(fractions & 0x0F) << (FRACTION_BITS-4);
      sp->loop_end |=
	((fractions>>4) & 0x0F) << (FRACTION_BITS-4);
    }

  close_file(fp);
  return ip;
}

static int fill_bank(int dr, int b)
{
  int i, errors=0;
  ToneBank *bank=((dr) ? drumset[b] : tonebank[b]);
  if (!bank)
    {
      cmsg(CMSG_ERROR, VERB_NORMAL, 
	   "Huh. Tried to load instruments in non-existent %s %d",
	   (dr) ? "drumset" : "tone bank", b);
      return 0;
    }
  for (i=0; i<128; i++)
    {
      if (bank->tone[i].instrument==MAGIC_LOAD_INSTRUMENT)
	{
	  if (!(bank->tone[i].name))
	    {
	      cmsg(CMSG_WARNING, VERB_NORMAL, 
		   "No instrument mapped to %s %d, program %d%s",
		   (dr)? "drum set" : "tone bank", b, i, 
		   (b!=0) ? "" : " - this instrument will not be heard");
	      if (b!=0)
		{
		  /* Mark the corresponding instrument in the default
		     bank / drumset for loading (if it isn't already) */
		  if (!dr)
		    {
		      if (!(standard_tonebank.tone[i].instrument))
			standard_tonebank.tone[i].instrument=
			  MAGIC_LOAD_INSTRUMENT;
		    }
		  else
		    {
		      if (!(standard_drumset.tone[i].instrument))
			standard_drumset.tone[i].instrument=
			  MAGIC_LOAD_INSTRUMENT;
		    }
		}
	      bank->tone[i].instrument=0;
	      errors++;
	    }
	  else if (!(bank->tone[i].instrument=
		     load_instrument(bank->tone[i].name)))
	    {
	      cmsg(CMSG_ERROR, VERB_NORMAL, 
		   "Couldn't load instrument %s (%s %d, program %d)",
		   bank->tone[i].name,
		   (dr)? "drum set" : "tone bank", b, i);
	      errors++;
	    }
	}
    }
  return errors;
}

int load_missing_instruments(void)
{
  int i=128,errors=0;
  while (i--)
    {
      if (tonebank[i])
	errors+=fill_bank(0,i);
      if (drumset[i])
	errors+=fill_bank(1,i);
    }
  return errors;
}

void free_instruments(void)
{
  int i=128;
  while(i--)
    {
      if (tonebank[i])
	free_bank(0,i);
      if (drumset[i])
	free_bank(1,i);
    }
}

int set_default_instrument(char *name)
{
  Instrument *ip;
  if (!(ip=load_instrument(name)))
    return -1;
  if (default_instrument)
    free_instrument(default_instrument);
  default_instrument=ip;
  default_program=SPECIAL_PROGRAM;
  return 0;
}
