/*

    TiMidity -- Experimental MIDI to WAVE converter
    Copyright (C) 1995 Tuukka Toivonen <titoivon@snakemail.hut.fi>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <stdio.h>
#include <unistd.h>
#include <strings.h>

#include "config.h"
#include "common.h"
#include "instruments.h"
#include "playmidi.h"
#include "readmidi.h"
#include "audio.h"
#include "controls.h"

#include "vol_table.h"
#include "freq_table.h"
#include "bend_table.h"

Channel channel[16];
Voice voice[MAX_VOICES];
int
    voices=DEFAULT_VOICES, control_ratio=0,
    amplification=DEFAULT_AMPLIFICATION, amp_with_poly;

static int lost_notes, cut_notes;
static int common_buffer[AUDIO_BUFFER_SIZE*2]; /* 2 ints per stereo sample */
static int buffered_count, *buffer_pointer;
static short resample_buffer[AUDIO_BUFFER_SIZE];

static MidiEvent *event_list, *current_event;
static int sample_count, current_sample;

int dump_midi(MidiEvent *event, int events)
{
  int i;
  for (i=0; i<events; i++)
    {
      switch(event[i].type)
	{
	  
	case ME_NOTEON:
	  fprintf(msgfp,
		  "%8ld: [%2d] %s on: note %d, velocity %3d\n",
		  event[i].time, event[i].channel+1,
		  (ISDRUMCHANNEL(event[i].channel)) ? "Drum" : "Note",
		  event[i].a, event[i].b);
	  break;
	  
	case ME_NOTEOFF:
	  fprintf(msgfp, "%8ld: [%2d] %s off: note %d, velocity %3d\n",
		  event[i].time, event[i].channel+1,
		  (ISDRUMCHANNEL(event[i].channel)) ? "Drum" : "Note",
		  event[i].a, event[i].b);
	  break;
	  
	case ME_KEYPRESSURE:
	  fprintf(msgfp, "%8ld: [%2d] Key pressure: note %d, pressure %3d\n",
		  event[i].time, event[i].channel+1,
		  event[i].a, event[i].b);
	  break;
	  
	case ME_MAINVOLUME:
	  fprintf(msgfp, "%8ld: [%2d] Main volume: %d\n",
		  event[i].time, event[i].channel+1,
		  event[i].a);
	  break;
	case ME_PAN:
	  fprintf(msgfp, "%8ld: [%2d] Panning: %d\n",
		  event[i].time, event[i].channel+1,
		  event[i].a);
	  break;

	case ME_SUSTAIN:
	  fprintf(msgfp, "%8ld: [%2d] Sustain: %s\n",
		  event[i].time, event[i].channel+1,
		  (event[i].a) ? "on" : "off");
	  break;
	  
	case ME_PITCHWHEEL:
	  fprintf(msgfp, "%8ld: [%2d] Pitch bend: %d\n",
		  event[i].time, event[i].channel+1,
		  event[i].a + event[i].b * 128);
	  break;

	case ME_EXPRESSION:
	  fprintf(msgfp, "%8ld: [%2d] Expression: %d\n",
		  event[i].time, event[i].channel+1,
		  event[i].a);
	  break;
	  
	case ME_PROGRAM:
	  fprintf(msgfp, "%8ld: [%2d] Program: %d\n",
		  event[i].time, event[i].channel+1,
		  event[i].a);
	  break;
	  
	case ME_TEMPO:
	  fprintf(msgfp, "%8ld: Tempo = %d\n",
		  event[i].time,
		  event[i].channel + event[i].b * 256 + event[i].a * 65536);
	  break;
	  
	case ME_NONE:
	default:
	  fprintf(stderr,
		  "*** Can't happen: Event %d in list!\n", event[i].type);
	  break;
	}
    }
  return 0;
}


static void adjust_amplification(void)
{ 
  amp_with_poly = (amplification * 65536) / (/*voices **/ 100);
}

static void all_notes_off(void)
{
  int i;
  for (i=0; i<MAX_VOICES; i++)
    voice[i].status=VOICE_FREE;
}

static void reset_midi(void)
{
  int i;
  for (i=0; i<16; i++)
    {
      channel[i].volume=90; /* Some standard says. */
      channel[i].expression=64;
      channel[i].sustain=0;
      channel[i].program=default_program;
      channel[i].panning=NO_PANNING;
      channel[i].pitchbend=0x2000;
      channel[i].bank=0; /* tone bank or drum set */
    }
  all_notes_off();
}

static void select_sample(int v, Instrument *ip)
{
  int s,i,f, cdiff, diff;
  Sample *sp, *closest;

  s=ip->samples;
  sp=ip->sample;

  if (s==1)
    voice[v].sample=sp;

  f=voice[v].orig_frequency;
  for (i=0; i<s; i++)
    {
      if (sp->low_freq <= f && sp->high_freq >= f)
	{
	  voice[v].sample=sp;
	  return;
	}
      sp++;
    }

  /* 
     No suitable sample found! This is a strangely frequent
     occurrence. Perhaps I'm doing something wrong.
     We'll select the sample whose root frequency is closest to the
     one we want. (Actually we should compare logarithms. We should
     probably convert the low, high, and root frequencies to MIDI note
     values and compare those.) */

  cdiff=0x7FFFFFFF;
  closest=0;
  sp=ip->sample;
  for(i=0; i<s; i++)
    {
      diff=sp->root_freq - f;
      if (diff<0) diff=-diff;
      if (diff<cdiff)
	{
	  cdiff=diff;
	  closest=sp;
	}
      sp++;
    }
  voice[v].sample=closest;
  return;
}

static void recompute_freq(int v)
{
  long long a,b;
  int pb=channel[voice[v].channel].pitchbend;
  int sign=(voice[v].sample_increment < 0);

  if (pb==0x2000)
    voice[v].frequency=voice[v].orig_frequency;
  else
    {
      if(pb<0x2000)
	{
	  a = (long long)(voice[v].orig_frequency)<<14;
	  a /= (long long)(bend_table[(0x1FFF-pb)>>6]);
	  voice[v].frequency=a;
	}
      else
	{
	  a = voice[v].orig_frequency;
	  a *= (long long)(bend_table[(pb-0x2000) >> 6]);
	  voice[v].frequency=(a >> 14);
	}
    }

  a = 
    (long long)(voice[v].sample->sample_rate) *
      (long long)(voice[v].frequency);
  b = 
    (long long)(voice[v].sample->root_freq) *
      (long long)(output_rate);

  a <<= FRACTION_BITS;
  a /= b;
  if (sign) a = -a; /* need to keep going where we were going */

  voice[v].sample_increment = (int)(a);
}

/* Returns 1 if envelope runs out */
static int recompute_envelope(int v)
{
  int stage;

  stage = voice[v].envelope_stage;

  if (stage>5)
    {
      /* Envelope ran out. */
      voice[v].status = VOICE_FREE;
      if(trace_playing)
	ctl_note(v);
      return 1;
    }

  if (((voice[v].sample->modes & (MODES_ENVELOPE|MODES_SUSTAIN))
       == (MODES_ENVELOPE|MODES_SUSTAIN)))
    {
      if (stage==1)
	stage = voice[v].envelope_stage = 2; 
      /* Skip "Attack 2" and go straight to "Sustain" */
      else
	if (voice[v].status==VOICE_ON || voice[v].status==VOICE_SUSTAINED)
	  {
	    if (stage>2)
	      {
		/* Freeze envelope until note turns off. Trumpets want this. */
		voice[v].envelope_increment=0;
		return 0;
	      }
	  }
    }
  voice[v].envelope_stage=stage+1;

  if (voice[v].envelope_volume==voice[v].sample->envelope_offset[stage])
    return recompute_envelope(v);
  voice[v].envelope_target=voice[v].sample->envelope_offset[stage];
  voice[v].envelope_increment = voice[v].sample->envelope_rate[stage];
  if (voice[v].envelope_target<voice[v].envelope_volume)
    voice[v].envelope_increment = -voice[v].envelope_increment;
  return 0;
}

/* alternatives: */
  /*      amp = velocity_table[voice[v].velocity] *	  
	  velocity_table[channel[voice[v].channel].volume]; */
  /*      amp = vol_table[voice[v].velocity] * 
	  (channel[voice[v].channel].volume << 5); */

static void recompute_amp(int v)
{
  int amp, left, right;

  amp = ((voice[v].velocity *
	  channel[voice[v].channel].volume) * amp_with_poly) >> 18;
  amp *= voice[v].sample->volume;
  amp >>= 4;
  amp *= channel[voice[v].channel].expression;
  amp >>= 6;
  
  if (!mono)
    {
      if (voice[v].panning > 60 && voice[v].panning < 68)
	{
	  left=32;
	  voice[v].panned = PANNED_CENTER;
	}
      else if (voice[v].panning<5)
	{
	  left=63; /* Right volume is ignored */
	  voice[v].panned = PANNED_LEFT;
	}
      else if (voice[v].panning>123)
	{
	  left=63; /* Right volume is still ignored. */
	  voice[v].panned = PANNED_RIGHT;
	}
      else
	{
	  right=voice[v].panning>>1;
	  left=63-right;
	  voice[v].panned = PANNED_MYSTERY;
	  /* Need the right volume as well */
	  voice[v].right_amp=(right * amp)>>16;
	}
    }
  else
    {
      left=32;
      voice[v].panned=PANNED_CENTER; /* It's mono output anyway  */
    }
  voice[v].left_amp=(left * amp)>>16; /* 14 or more bits */
}

/* Returns 1 if the note died */
int update_signal(int v)
{
  voice[v].envelope_volume += voice[v].envelope_increment;
  
  /* Why is there no ^^ operator?? */
  if (((voice[v].envelope_increment < 0) &&
       (voice[v].envelope_volume <= voice[v].envelope_target)) ||
      ((voice[v].envelope_increment > 0) &&
       (voice[v].envelope_volume >= voice[v].envelope_target)))
    {
      voice[v].envelope_volume = voice[v].envelope_target;
      if (recompute_envelope(v))
	return 1;
    }
  voice[v].left_envelope=
    (voice[v].left_amp * ((vol_table[voice[v].envelope_volume>>23])))>>12;
  
  if (voice[v].panned == PANNED_MYSTERY) /* need right envelope as well */
    voice[v].right_envelope=
      (voice[v].right_amp * ((vol_table[voice[v].envelope_volume>>23])))>>12;

  return 0;
}

static void start_note(MidiEvent *e, int i)
{
  Instrument *ip;

  if (ISDRUMCHANNEL(e->channel))
    {
      if (!(ip=drumset[channel[e->channel].bank]->tone[e->a].instrument))
	{
	  if (!(ip=drumset[0]->tone[e->a].instrument))
	    return; /* No instrument? Then we can't play. */
	}
    }
  else
    {
      if (channel[e->channel].program==SPECIAL_PROGRAM)
	ip=default_instrument;
      else if (!(ip=tonebank[channel[e->channel].bank]->
		 tone[channel[e->channel].program].instrument))
	{
	  if (!(ip=tonebank[0]->tone[channel[e->channel].program].instrument))
	    return; /* No instrument? Then we can't play. */
	}
    }

  voice[i].orig_frequency=freq_table[e->a & 0x7F];
  select_sample(i, ip);

  voice[i].status=VOICE_ON;
  voice[i].channel=e->channel;
  voice[i].note=e->a;
  voice[i].velocity=e->b;
  voice[i].sample_offset=0;
  voice[i].sample_increment=0; /* make sure it isn't negative */

  if (channel[e->channel].panning != NO_PANNING)
    voice[i].panning=channel[e->channel].panning;
  else
    voice[i].panning=voice[i].sample->panning;

  recompute_freq(i);
  recompute_amp(i);
  if (voice[i].sample->modes & MODES_ENVELOPE)
    {
      /*int tmp;*/
      /* Ramp up from 0 */
      voice[i].envelope_stage=0;
      voice[i].envelope_volume=0;
      voice[i].control_counter=0;
      recompute_envelope(i);
      /* just testing */
      /*tmp=voice[i].envelope_volume;
      voice[i].envelope_volume=0;*/
      update_signal(i); /* We need valid data in left_envelope. */
      /*voice[i].envelope_volume=tmp;*/
    }
  else
    {
      voice[i].left_envelope=voice[i].left_amp;
      voice[i].right_envelope=voice[i].right_amp;
    }
  if (trace_playing)
    ctl_note(i);
}

/* Only one instance of a note can be playing on a single channel. */
static void note_on(MidiEvent *e)
{
  int i=voices, lowest=-1, lv=0x7FFFFFFF, v;

  while (i--)
    if (voice[i].status == VOICE_FREE)
      lowest=i; /* Can't get a lower volume than silence */
    else if (voice[i].channel==e->channel && voice[i].note==e->a)
      {
#ifndef FOWL_PLAY_CLICK
	start_note(e,i);
	return;
#else
	/* Kill this note. We can't just start a new note over this 
	   one, as that seems to cause snaps. */
	/* Do this right some day! */
	voice[i].status=VOICE_OFF;
	voice[i].envelope_stage=6;
	voice[i].envelope_target=0;
	voice[i].envelope_increment=-20000000;
	/* Yeah... Retriggered notes disappear from the display. Bummer. */
	if (trace_playing)
	  ctl_note(i);
	update_signal(i);
#endif
      }

  /* Found a free voice to play on? */
  if (lowest != -1)
    {
      voice[lowest].status=VOICE_FREE;
      if (trace_playing)
	ctl_note(lowest);
      start_note(e,lowest);
      return;
    }
  
  i=voices;
  while (i--)
    {
      if ((voice[i].status!=VOICE_ON) &&
	  (voice[i].sample->modes & MODES_ENVELOPE))
	{
	  v=voice[i].envelope_volume;
	  if (v<lv)
	    {
	      lv=v;
	      lowest=i;
	    }
	}
    }

  /* Found a decaying voice to overtake? */
  if (lowest != -1)
    {
      voice[lowest].status=VOICE_FREE;
      if (trace_playing)
	ctl_note(lowest);
      cut_notes++;
      start_note(e,lowest);
      return;
    }
  
  /* Couldn't start that note */
  lost_notes++;
}

static void finish_note(int i)
{
  if ((voice[i].sample->modes & (MODES_ENVELOPE|MODES_SUSTAIN))==
      (MODES_ENVELOPE|MODES_SUSTAIN))
    {
      /* We need to get the envelope out of Sustain mode */
      voice[i].envelope_stage=3;
      voice[i].status=VOICE_OFF;
      recompute_envelope(i);
      if (trace_playing)
	ctl_note(i);
    }
  else
    {
      if (voice[i].sample->modes & MODES_SUSTAIN)
	{
	  voice[i].status=VOICE_FREE;
	  if (trace_playing)
	    ctl_note(i);
	}
      else
	{
	  if (voice[i].sample->modes & MODES_LOOPING)
	    {
	      voice[i].status=VOICE_OFF;
	      if (trace_playing)
		ctl_note(i);
	    }
	  /* else play the entire sample in its entirety */
	}
    }
}

static void note_off(MidiEvent *e)
{
  int i=voices;
  while (i--)
    if (voice[i].status==VOICE_ON &&
	voice[i].channel==e->channel &&
	voice[i].note==e->a)
      {
	if (channel[e->channel].sustain)
	  {
	    voice[i].status=VOICE_SUSTAINED;
	    if (trace_playing)
	      ctl_note(i);
	  }
	else
	  finish_note(i);
	return;
      }
}

static void adjust_pressure(MidiEvent *e)
{
  int i=voices;
  while (i--)
    if (voice[i].status==VOICE_ON &&
	voice[i].channel==e->channel &&
	voice[i].note==e->a)
      {
	voice[i].velocity=e->b;
	recompute_amp(i);
	update_signal(i);
	return;
      }
}

static void adjust_panning(int c)
{
  int i=voices;
  while (i--)
    if (voice[i].status!=VOICE_FREE &&
	voice[i].channel==c)
      {
	voice[i].panning=channel[c].panning;
	recompute_amp(i);
	update_signal(i);
	return;
      }
}

static void drop_sustain(int c)
{
  int i=voices;
  while (i--)
    if (voice[i].status==VOICE_SUSTAINED && voice[i].channel==c)
      finish_note(i);
}

static void adjust_pitchbend(int c)
{
  int i=voices;
  while (i--)
    if (voice[i].status!=VOICE_FREE && voice[i].channel==c)
      recompute_freq(i);
}

static void adjust_volume(int c)
{
  int i=voices;
  while (i--)
    if (voice[i].status!=VOICE_FREE && voice[i].channel==c)
      {
	recompute_amp(i);
	update_signal(i);
      }
}

  /*
     Slight overflow of the common buffer (into the guard bits) 
     will be clipped without serious distortion in the output stage.

     Still, it's possible to get a combination of volumes, envelopes,
     and amplification that'll overflow the guard bits as well, and
     the output samples will roll around, causing terrible tearing
     noises.

     There's a lot of duplicated code here to take advantage of the fact
     that many instruments are panned at center. This saves us one
     multiplication per mixation. (cf. resamplation)
   */
  
void mix_data(int v, int count)
{
  int left, right, cc;
  int *lp=buffer_pointer;
  short *sp=resample_buffer, s;

  left=voice[v].left_envelope;
  if (!mono)
    {
      right=voice[v].right_envelope;
      if (voice[v].panned==PANNED_MYSTERY)
	{
	  if ((voice[v].sample->modes & MODES_ENVELOPE) &&
	      voice[v].envelope_increment)
	    {
	      cc=voice[v].control_counter;
	      while (count--)
		{
		  if (!cc--)
		    {
		      cc=control_ratio;
		      if (update_signal(v))
			return; /* Envelope ran out */
		      left=voice[v].left_envelope;
		      right=voice[v].right_envelope;
		    }
		  s=*sp++;
		  *lp++ += left * s;
		  *lp++ += right * s;
		}
	      voice[v].control_counter=cc;
	    }
	  else /* No envelope! This is too easy... */
	    {
	      while (count--)
		{
		  s=*sp++;
		  *lp++ += left * s;
		  *lp++ += right * s;
		}	  
	    }
	}
      else if (voice[v].panned==PANNED_CENTER)
	{
	  if (voice[v].sample->modes & MODES_ENVELOPE)
	    {
	      cc=voice[v].control_counter;
	      while (count--)
		{
		  if (!cc--)
		    {
		      cc=control_ratio;
		      if (update_signal(v))
			return; /* Envelope ran out */
		      left=voice[v].left_envelope;
		    }
		  s=*sp++;
		  *lp++ += left * s;
		  *lp++ += left * s;
		}
	      voice[v].control_counter=cc;
	    }
	  else /* No envelope! This is too easy... */
	    {
	      while (count--)
		{
		  s=*sp++;
		  *lp++ += left * s;
		  *lp++ += left * s;
		}	  
	    }
	}
      else if (voice[v].panned==PANNED_LEFT)
	{
	  if (voice[v].sample->modes & MODES_ENVELOPE)
	    {
	      cc=voice[v].control_counter;
	      while (count--)
		{
		  if (!cc--)
		    {
		      cc=control_ratio;
		      if (update_signal(v))
			return; /* Envelope ran out */
		      left=voice[v].left_envelope;
		    }
		  s=*sp++;
		  *lp++ += left * s;
		  lp++;
		}
	      voice[v].control_counter=cc;
	    }
	  else /* No envelope! This is too easy... */
	    {
	      while (count--)
		{
		  s=*sp++;
		  *lp++ += left * s;
		  lp++;
		}	  
	    }
	}
      else if (voice[v].panned==PANNED_RIGHT)
	{
	  if (voice[v].sample->modes & MODES_ENVELOPE)
	    {
	      cc=voice[v].control_counter;
	      while (count--)
		{
		  if (!cc--)
		    {
		      cc=control_ratio;
		      if (update_signal(v))
			return; /* Envelope ran out */
		      left=voice[v].left_envelope;
		    }
		  s=*sp++;
		  lp++;
		  *lp++ += left * s;
		}
	      voice[v].control_counter=cc;
	    }
	  else /* No envelope! This is too easy... */
	    {
	      while (count--)
		{
		  s=*sp++;
		  lp++;
		  *lp++ += left * s;
		}	  
	    }
	}
    }
  else
    {
      /* Mono output.  */
      if (voice[v].sample->modes & MODES_ENVELOPE)
	{
	  cc=voice[v].control_counter;
	  while (count--)
	    {
	      if (!cc--)
		{
		  cc=control_ratio;
		  if (update_signal(v))
		    return; /* Envelope ran out */
		  left=voice[v].left_envelope;
		}
	      s=*sp++;
	      *lp++ += left * s;
	    }
	  voice[v].control_counter=cc;
	}
      else /* No envelope! This is too easy... */
	{
	  while (count--)
	    {
	      s=*sp++;
	      *lp++ += left * s;
	    }	  
	}
    }
}

#ifdef LINEAR_INTERPOLATION
/* Addition, division, interpolation. Resamplation. */
#define RESAMPLATION \
      v1=src[ofs>>FRACTION_BITS];\
      v2=src[(ofs>>FRACTION_BITS)+1];\
      *dest++ = v1 + (((v2-v1) * (ofs & FRACTION_MASK)) >> FRACTION_BITS);
#else
#define RESAMPLATION \
      *dest++=src[ofs>>FRACTION_BITS]; /* Sounds terrible */
#endif

/* It might be worthwhile to combine the resampling and mixing stages,
   since Intel chips can't even hold one stage in registers at a time,
   we're constantly spilling anyway, and that way you'd save a memory
   write and read for each sample. Assembly could help. */
int resample_data(int v, int count)
{
  unsigned long ofs, incr, le, ll, ls;
  int orig_count=count;
  Voice *vp;
  short *dest,*src;
#ifdef LINEAR_INTERPOLATION
  short v1, v2;
#endif
  unsigned char modes;

  vp=&voice[v];

  src=vp->sample->data;
  dest=resample_buffer;

  incr=vp->sample_increment;
  ofs=vp->sample_offset;

  modes=vp->sample->modes;

  if (!(modes & MODES_LOOPING) ||
      (!(modes & (MODES_SUSTAIN | MODES_ENVELOPE)) && 
       vp->status == VOICE_OFF))
    {
      /* Simple: Play sample until end, then free the voice.
	 (That's what I'd do, but some samples have a >snap!< at the end,
	 so maybe I'm supposed to play until loop end instead. Weird!) */
      /*le=vp->sample->data_length;*/
      le=vp->sample->loop_end;
      while (count--)
	{
	  ofs += incr;
	  if (ofs >= le)
	    {
	      vp->status=VOICE_FREE; /* This sound is OVER. */
	      if (trace_playing)
		ctl_note(v);
	      return (orig_count-count-1);
	    }
	  RESAMPLATION;
	}
    }
  else if (!(modes & MODES_PINGPONG))
    {
      /* Simple: Play sample until end-of-loop, then skip backwards a bit
	 and continue. */
      le=vp->sample->loop_end;
      ll=le - vp->sample->loop_start;
      while (count--)
	{
	  ofs += incr;
	  if (ofs>=le)
	    ofs -= ll; /* Hopefully the loop is longer than an increment. */
	  RESAMPLATION;
	}
    }
  else
    {
      /* Bidirectional looping. */
      le=vp->sample->loop_end;
      ls=vp->sample->loop_start;
      if (ofs < ls) /* Play normally until inside the loop region */
	{
	  while (count--)
	    {
	      ofs += incr;
	      RESAMPLATION;
	      if (ofs>=ls)
		break; /* out of this loop and into the next */
	    }
	}

      if (count>0) /* Then do the bidirectional looping */
	while (count--)
	  {
	    ofs += incr;
	    if (ofs>=le)
	      {
		/* fold the overshoot back in */
		ofs = le - (ofs - le);
		incr = -incr;
	      }
	    else if (ofs <= ls)
	      {
		ofs = ls + (ls - ofs);
		incr = -incr;
	      }
	    RESAMPLATION;
	  }
      vp->sample_increment=incr;
    }
  vp->sample_offset=ofs; /* Update offset */
  return orig_count;
}

static void seek_forward(int until_time)
{
  all_notes_off();
  while (current_event->time < until_time)
    {
      switch(current_event->type)
	{
	  /* All notes stay off. Just handle the parameter changes. */
	      
	case ME_PITCHWHEEL:
	  channel[current_event->channel].pitchbend=
	    current_event->a + current_event->b * 128;
	  break;
	  
	case ME_MAINVOLUME:
	  channel[current_event->channel].volume=current_event->a;
	  break;
	  
	case ME_PAN:
	  channel[current_event->channel].panning=current_event->a;
	  break;
	      
	case ME_EXPRESSION:
	  channel[current_event->channel].expression=current_event->a;
	  break;
	  
	case ME_PROGRAM:
	  if (ISDRUMCHANNEL(current_event->channel))
	    /* Change drum set */
	    channel[current_event->channel].bank=current_event->a;
	  else
	    channel[current_event->channel].program=current_event->a;
	  break;

	case ME_SUSTAIN:
	  channel[current_event->channel].sustain=current_event->a;
	  break;
	      
	case ME_TONE_BANK:
	  channel[current_event->channel].bank=current_event->a;
	  break;
	  
	case ME_EOT:
	  current_sample=current_event->time;
	  return;
	}
      current_event++;
    }
  current_sample=current_event->time;
}

static void skip_to(int until_time)
{
  if (current_sample > until_time)
    current_sample=0;

  reset_midi();
  buffered_count=0;
  buffer_pointer=common_buffer;
  current_event=event_list;
  
  if (until_time)
    seek_forward(until_time);
  if (interactive)
    ctl_reset();
}

#ifdef INTERACTIVE
static int apply_controls()
{
  int rc, val, i, did_skip=0;
  /* ASCII renditions of CD player pictograms indicate approximate effect */
  do
    switch(rc=read_controls(&val))
      {
      case RC_QUIT: /* [] */
      case RC_NEXT: /* >>| */
      case RC_REALLY_PREVIOUS: /* |<< */
	purge_output();
	return rc;
	
      case RC_CHANGE_VOLUME:
	amplification += val;
	if (amplification < 0) amplification=0;
	else if (amplification > MAX_AMPLIFICATION)
	  amplification=MAX_AMPLIFICATION;
	adjust_amplification();
	for (i=0; i<voices; i++)
	  if (voice[i].status != VOICE_FREE)
	    {
	      recompute_amp(i);
	      update_signal(i);
	    }
	ctl_master_volume(amplification);
	break;

      case RC_PREVIOUS: /* |<< */
	purge_output();
	if (current_sample < 2*output_rate)
	  return RC_REALLY_PREVIOUS;
	return RC_RESTART;

      case RC_RESTART: /* |<< */
	skip_to(0);
	purge_output();
	return rc;
	
      case RC_JUMP:
	purge_output();
	if (val >= sample_count)
	  return RC_NEXT;
	skip_to(val);
	return rc;
	
      case RC_FORWARD: /* >> */
	/*purge_output();*/
	if (val+current_sample >= sample_count)
	  return RC_NEXT;
	skip_to(val+current_sample);
	did_skip=1;
	break;
	
      case RC_BACK: /* << */
	/*purge_output();*/
	if (current_sample > val)
	  skip_to(current_sample-val);
	else
	  skip_to(0); /* We can't seek to end of previous song. */
	did_skip=1;
	break;
      }
  while (rc!= RC_NONE);
  if (did_skip)
    return RC_JUMP; /* could have been RC_BACK or RC_FORWARD, not important */
  else
    return rc;
}
#endif 

static void do_compute_data(int count)
{
  int i;
  memset(buffer_pointer, 0, (mono) ? (count * 4) : (count * 8));
  for (i=0; i<voices; i++)
    {
      if(voice[i].status != VOICE_FREE)
	mix_data(i, resample_data(i, count)); 
    }
  current_sample += count;
}

/* count=0 means flush buffered data to device */
static int compute_data(int count)
{
  int rc;
  if (!count)
    {
      if (buffered_count)
	output_data(common_buffer, buffered_count);
      flush_output();
      buffer_pointer=common_buffer;
      buffered_count=0;
      return RC_NONE;
    }

  while ((count+buffered_count) >= AUDIO_BUFFER_SIZE)
    {
      do_compute_data(AUDIO_BUFFER_SIZE-buffered_count);
      count -= AUDIO_BUFFER_SIZE-buffered_count;
      output_data(common_buffer, AUDIO_BUFFER_SIZE);
      buffer_pointer=common_buffer;
      buffered_count=0;
      
      if (interactive)
	{
	  ctl_current_time(current_sample);
#ifdef INTERACTIVE
	  if ((rc=apply_controls())!=RC_NONE)
	    return rc;
#endif
	}
    }
  if (count>0)
    {
      do_compute_data(count);
      buffered_count += count;
      buffer_pointer += (mono) ? count : count*2;
    }
  return RC_NONE;
}

int play_midi(MidiEvent *eventlist, int events, int samples)
{
  int rc;

  adjust_amplification();

  sample_count=samples;
  event_list=eventlist;
  lost_notes=cut_notes=0;

  skip_to(0);
  
  for (;;)
    {
      /* Handle all events that should happen at this time */
      while (current_event->time <= current_sample)
	{
	  switch(current_event->type)
	    {

	      /* Effects affecting a single note */

	    case ME_NOTEON:
	      if (!(current_event->b)) /* Velocity 0? */
		note_off(current_event);
	      else
		note_on(current_event);
	      break;

	    case ME_NOTEOFF:
	      note_off(current_event);
	      break;

	    case ME_KEYPRESSURE:
	      adjust_pressure(current_event);
	      break;

	      /* Effects affecting a single channel */
	      
	    case ME_PITCHWHEEL:
	      channel[current_event->channel].pitchbend=
		current_event->a + current_event->b * 128;
	      /* Adjust pitch for notes already playing */
	      adjust_pitchbend(current_event->channel);
	      if (interactive)
		ctl_pitch_bend(current_event->channel, 
			       channel[current_event->channel].pitchbend);
	      break;
	      
	    case ME_MAINVOLUME:
	      channel[current_event->channel].volume=current_event->a;
	      adjust_volume(current_event->channel);
	      if (interactive)
		ctl_volume(current_event->channel, current_event->a);
	      break;
	      
	    case ME_PAN:
	      channel[current_event->channel].panning=current_event->a;
	      adjust_panning(current_event->channel);
	      if (interactive)
		ctl_panning(current_event->channel, current_event->a);
	      break;
	      
	    case ME_EXPRESSION:
	      channel[current_event->channel].expression=current_event->a;
	      adjust_volume(current_event->channel);
	      if (interactive)
		ctl_expression(current_event->channel, current_event->a);
	      break;

	    case ME_PROGRAM:
	      if (ISDRUMCHANNEL(current_event->channel))
		{
		  /* Change drum set */
		  channel[current_event->channel].bank=current_event->a;
		}
	      else
		{
		  channel[current_event->channel].program=current_event->a;
		}
	      if (interactive)
		ctl_program(current_event->channel, current_event->a);
	      break;

	    case ME_SUSTAIN:
	      channel[current_event->channel].sustain=current_event->a;
	      if (!current_event->a)
		drop_sustain(current_event->channel);
	      if (interactive)
		ctl_sustain(current_event->channel, current_event->a);
	      break;
	      
	    case ME_TONE_BANK:
	      channel[current_event->channel].bank=current_event->a;
	      break;

	    case ME_EOT:
	      /* Give the last notes a couple of seconds to decay  */
	      compute_data(output_rate * 2);
	      compute_data(0); /* flush buffer to device */
	      cmsg(CMSG_INFO, VERB_VERBOSE,
		   "Playing time: ~%d seconds",
		   current_sample/output_rate+2);
	      cmsg(CMSG_INFO, VERB_VERBOSE,
		   "Notes cut: %d", cut_notes);
	      cmsg(CMSG_INFO, VERB_VERBOSE,
		   "Notes lost totally: %d", lost_notes);
	      return RC_NEXT;
	    }
	  current_event++;
	}

      rc=compute_data(current_event->time - current_sample);
      /*current_sample=current_event->time;*/
      if (interactive)
	{
	  ctl_refresh();
	  switch(rc)
	    {
	    case RC_QUIT:
	    case RC_NEXT:
	    case RC_REALLY_PREVIOUS:
	      return rc;
	    }
	}
    }
}
