/*

    TiMidity -- Experimental MIDI to WAVE converter
    Copyright (C) 1995 Tuukka Toivonen <titoivon@snakemail.hut.fi>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <strings.h>

#include "config.h"
#include "common.h"
#include "instruments.h"
#include "playmidi.h"
#include "readmidi.h"
#include "audio.h"
#include "controls.h"

static int free_instruments_afterwards=0;
static char def_instr_name[256]="";

static int map_patch_to_tone_bank(ToneBank *bank, int i, char *name)
{
  if (i<0 || i>127) return 1;
  if (bank->tone[i].name)
    free(bank->tone[i].name);
  strcpy((bank->tone[i].name=safe_malloc(strlen(name)+1)),name);
  return 0;
}

static int read_config_file(char *name)
{
  FILE *fp;
  char tmp[1024], *w1, *w2;
  ToneBank *bank=0;
  int i, line=0;
  static int rcf_count=0;

  if (rcf_count>50)
    {
      cmsg(CMSG_ERROR, VERB_NORMAL, 
	   "Probable endless loop in configuration files");
      return (-1);
    }

  cmsg(CMSG_INFO, VERB_DEBUG, "CONFIG FILE: %s", name);

  if (!(fp=open_file(name, 1, OF_VERBOSE)))
    return -1;

  while (fgets(tmp, sizeof(tmp), fp))
    {
      line++;
      w1=strtok(tmp, " \t\r\n\240");
      w2=strtok(0," \t\r\n\240");
      if (!w1 || (*w1=='#')) continue;
      else if (!strcmp(w1, "dir"))
	{
	  if (!w2)
	    {
	      cmsg(CMSG_ERROR, VERB_NORMAL, 
		   "%s: line %d: No directory given", name, line);
	      return -2;
	    }
	  add_to_pathlist(w2);
	}
      else if (!strcmp(w1, "source"))
	{
	  if (!w2)
	    {
	      cmsg(CMSG_ERROR, VERB_NORMAL, 
		   "%s: line %d: No file name given", name, line);
	      return -2;
	    }
	  rcf_count++;
	  read_config_file(w2);
	  rcf_count--;
	}
      else if (!strcmp(w1, "default"))
	{
	  if (!w2)
	    {
	      cmsg(CMSG_ERROR, VERB_NORMAL, 
		   "%s: line %d: No patch name given", name, line);
	      return -2;
	    }
	  strncpy(def_instr_name, w2, 255);
	  def_instr_name[255]='\0';
	}
      else if (!strcmp(w1, "drumset"))
	{
	  if (!w2)
	    {
	      cmsg(CMSG_ERROR, VERB_NORMAL,
		   "%s: line %d: No drum set number given", name, line);
	      return -2;
	    }
	  i=atoi(w2);
	  if (i<0 || i>127)
	    {
	      cmsg(CMSG_ERROR, VERB_NORMAL, 
		   "%s: line %d: Drum set must be between 0 and 127",
		   name, line);
	      return -2;
	    }
	  if (!drumset[i])
	    {
	      drumset[i]=safe_malloc(sizeof(ToneBank));
	      memset(drumset[i], 0, sizeof(ToneBank));
	    }
	  bank=drumset[i];
	}
      else if (!strcmp(w1, "bank"))
	{
	  if (!w2)
	    {
	      cmsg(CMSG_ERROR, VERB_NORMAL, 
		   "%s: line %d: No bank number given", name, line);
	      return -2;
	    }
	  i=atoi(w2);
	  if (i<0 || i>127)
	    {
	      cmsg(CMSG_ERROR, VERB_NORMAL, 
		   "%s: line %d: Tone bank must be between 0 and 127",
		   name, line);
	      return -2;
	    }
	  if (!tonebank[i])
	    {
	      tonebank[i]=safe_malloc(sizeof(ToneBank));
	      memset(tonebank[i], 0, sizeof(ToneBank));
	    }
	  bank=tonebank[i];
	}
      else {
	if (!w2 || (*w1 < '0' || *w1 > '9'))
	  {
	    cmsg(CMSG_ERROR, VERB_NORMAL, 
		 "%s: line %d: syntax error", name, line);
	    return -2;
	  }
	i=atoi(w1);
	if (i<0 || i>127)
	  {
	    cmsg(CMSG_ERROR, VERB_NORMAL, 
		 "%s: line %d: Program must be between 0 and 127",
		 name, line);
	    return -2;
	  }
	map_patch_to_tone_bank(bank, i, w2);
      }
    }
  close_file(fp);
  return 0;
}

static int dumpmidi=0;

static int play_midi_file(char *fn)
{
  MidiEvent *event;
  int events, samples, rc;
  FILE *fp;

  cmsg(CMSG_INFO, VERB_NORMAL, "Playing file: %s", fn);

  if (!strcmp(fn, "-"))
    {
      fp=stdin;
      strcpy(current_filename, "(stdin)");
    }
  else if (!(fp=open_file(fn, 1, OF_VERBOSE)))
    return -1;

  if (interactive) 
    ctl_file_name(current_filename);

  event=read_midi_file(fp, &events, &samples);

  if (fp != stdin)
    close_file(fp);

  if (!event)
    return -1;

  cmsg(CMSG_INFO, VERB_NOISY, 
       "%d supported events, %d samples", events, samples);

  if (interactive)
    {
      ctl_total_time(samples);
      ctl_master_volume(amplification);
    }
  else if (dumpmidi)
    dump_midi(event, events);

  
  load_missing_instruments();

  rc=play_midi(event, events, samples);

  free(event);
  if (free_instruments_afterwards)
    free_instruments();
  return rc;
}

int main(int argc, char **argv)
{
  int c, cmderr=0, i, got_a_configuration=0, try_config_again=0;
  if ((program_name=rindex(argv[0], '/'))) program_name++;
  else program_name=argv[0];
  if (argc==1) 
    {
      printf(
"\n"
" TiMidity version " TIMID_VERSION " -- Experimental MIDI to WAVE converter\n"
" Copyright (C) 1995 Tuukka Toivonen <titoivon@snakemail.hut.fi>\n"
" \n"
" This program is free software; you can redistribute it and/or modify\n"
" it under the terms of the GNU General Public License as published by\n"
" the Free Software Foundation; either version 2 of the License, or\n"
" (at your option) any later version.\n"
" \n"
" This program is distributed in the hope that it will be useful,\n"
" but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
" MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
" GNU General Public License for more details.\n"
" \n"
" You should have received a copy of the GNU General Public License\n"
" along with this program; if not, write to the Free Software\n"
" Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n"
"\n"
);
      return 0;
    }

  if (!read_config_file(CONFIG_FILE))
    got_a_configuration=1;

  while ((c=getopt(argc, argv, "UqvI:8L:c:A:P:C:d:Mp:o:W:s:hV"
#ifdef INTERACTIVE
		   "iB:"
#endif
		   ))>0)
    switch(c)
      {
      case 'U': free_instruments_afterwards=1; break;
      case 'q':	verbosity--; break;
      case 'v':	verbosity++; break;
      case 'V': dumpmidi=1; break;
      case 'I':
	i=atoi(optarg);
	if (i<0 || i>127)
	  {
	    cmsg(CMSG_ERROR, VERB_NORMAL, 
		 "Default program must be between 0 and 127");
	    cmderr++;
	  }
	else default_program=i;
	break;
      case '8': eight_bit=1; break;
      case 'L': add_to_pathlist(optarg); try_config_again=1; break;
      case 'c':
	if (read_config_file(optarg)) cmderr++;
	else got_a_configuration=1;
	break;

      case 'A':
	i=atoi(optarg);
	if (i<1 || i > MAX_AMPLIFICATION)
	  {
	    cmsg(CMSG_ERROR, VERB_NORMAL, 
		 "Amplification must be between 1 and %d %%",
		 MAX_AMPLIFICATION);
	    cmderr++;
	  }
	else amplification=i;
	break;
	
      case 'P':
	strncpy(def_instr_name, optarg, 255);
	def_instr_name[255]='\0';
	break;

      case 'C':
	i=atoi(optarg);
	if (i<1 || i > MAX_CONTROL_RATIO)
	  {
	    cmsg(CMSG_ERROR, VERB_NORMAL, 
		 "Control ratio must be between 1 and %d",
		 MAX_CONTROL_RATIO);
	    cmderr++;
	  }
	else control_ratio=i;
	break;

      case 'd': audio_name=optarg; play_mode=PLAY_DEVICE; break;
      case 'M': mono=1; break;
      case 'p':
	i=atoi(optarg);
	if (i < 1 || i > MAX_VOICES)
	  {
	    cmsg(CMSG_ERROR, VERB_NORMAL, 
		 "Polyphony must be between 1 and %d", MAX_VOICES);
	    cmderr++;
	  }
	else voices=i;
	break;

      case 'o':
	play_mode=PLAY_FILE;
	interactive=trace_playing=0;
	audio_name=optarg;
	break;

      case 'W':
	play_mode=PLAY_WAVE;
	interactive=trace_playing=0;
	audio_name=optarg;
	break;

#ifdef INTERACTIVE
      case 'i':
	if (play_mode == PLAY_DEVICE)
	  {
	    interactive=trace_playing=1;
	    if (buffer_fragments==-1)
	      buffer_fragments=2;
	  }
	break;

      case 'B': 
	i=atoi(optarg);
	if (i!=0 && (i<2 || i>1000))
	  {
	    cmsg(CMSG_ERROR, VERB_NORMAL, 
		 "Buffer fragments must be 0 or between 2 and 1000");
	    cmderr++;
	  }
	else buffer_fragments=i;
	break;
#endif

      case 's':
	i=atoi(optarg);
	if (i < 100)
	  i *= 1000;
	if (i<MIN_OUTPUT_RATE || i>MAX_OUTPUT_RATE)
	  {
	    cmsg(CMSG_ERROR, VERB_NORMAL, 
		 "Output frequency should be between %d and %d Hz.", 
		 MIN_OUTPUT_RATE, MAX_OUTPUT_RATE);
	    cmderr++;
	  }
	else output_rate=i;
	break;

      case 'h':
	printf(" TiMidity version " TIMID_VERSION " (C) 1995 Tuukka Toivonen "
		"<titoivon@snakemail.hut.fi>\n"
	       " TiMidity is free software and comes with "
		"ABSOLUTELY NO WARRANTY.\n"
	       "\n"
	       "Usage:\n"
	       "  %s [options] filename [...]\n"
	       "\n"
	       "  Use \"-\" as filename to read a MIDI file from stdin\n"
	       "\n"
	       "Options:\n"
	       "  -s f    Set sampling frequency to f (Hz or kHz)\n"
	       "  -M      Generate monophonic output\n"
	       "  -8      Generate 8-bit output\n"
	       "  -d dev  Play on audio device dev\n"
	       "  -o file Play to file (Use \"-\" to play to stdout)\n"
	       "  -W file Play to WAVE file\n"
	       "  -p n    Allow n-voice polyphony\n"
	       "  -A n    Amplify volume by n percent (may cause clipping)\n"
	       "  -C n    Set ratio of sampling and control frequencies\n"
	       "\n"
	       "  -L dir  Append dir to search path\n"
	       "  -c file Read extra config file\n"
	       "  -I n    Use program n as the default\n"
	       "  -P file Use patch file for all programs\n"
	       "  -U      Unload instruments from memory between MIDI files\n"
	       "\n"
	       "  -vv...  Increase verbosity\n"
	       "  -qq...  Decrease verbosity\n"
	       /*"  -t      Trace playing\n"*/
#ifdef INTERACTIVE
	       "  -i      Interactive mode\n"
	       "  -B n    Set number of buffer fragments in interactive mode\n"
#endif
	       "  -h      Display this help message\n"
	       "\n"
		, program_name);
	return 0;
      default:
	cmderr++; break;
      }

  if (!got_a_configuration && try_config_again && 
      !read_config_file(CONFIG_FILE))
    cmderr++;

  if (cmderr || optind >= argc)
    {
      cmsg(CMSG_ERROR, VERB_NORMAL, 
	   "Try %s -h for help", program_name);
      return 1; /* problems with command line */
    }

  if (!(isatty(fileno(inputfp))))
    {
      if (isatty(fileno(msgfp)))
	inputfp=msgfp;
      else if (isatty(fileno(stderr)))
	inputfp=stderr;
      else
	interactive=trace_playing=0;
    }

  if (optind<argc)
    {
      int orig_optind=optind;
#ifdef INTERACTIVE
      if (interactive)
	open_controls();
#endif
      if (open_output())
	bail_out(11); /* this could be done better */

      if (*def_instr_name)
	set_default_instrument(def_instr_name);

      for (;;)
	{
	  switch(play_midi_file(argv[optind]))
	    {
	    case RC_REALLY_PREVIOUS:
	      if (optind>orig_optind)
		optind--;
	      break;

	    default: /* Could have been an error or something */
	    case RC_NEXT:
	      if (optind<(argc-1))
		{
		  optind++;
		  break;
		}
	      /* else FALL THROUGH */

	    case RC_QUIT:
	      close_output();
#ifdef INTERACTIVE
	      if (interactive)
		close_controls();
#endif
	      return 0;

	    }
	}
    }
  return 0;
}
